enum matrixType { FULL, BLOCK, SYMMETRIC, COLUMNCOMPACTED };
#define TRUE 1
#define FALSE 0

/****************************/
/*       FeaVector          */
/****************************/
class FeaFullVector;
class FeaVector {
	matrixType type;
public:
	FeaVector(matrixType t) { type = t; };
	virtual ~FeaVector(){};
	void operator=(const FeaVector& other) { type = other.type; };
	FeaVector(const FeaVector& other){ type = other.type; };
	int operator==(matrixType t){return(type == t);};

	virtual double& operator()(int)=0;
	virtual FeaFullVector& getSubVector(int, int)=0;

	virtual void print()=0;
	virtual void printTerse()=0;
};

class FeaMatrix;
class FeaFullMatrix;
/****************************/
/*     FeaFullVector        */
/****************************/

class FeaFullVector : public FeaVector{
	int size;
	double* data;
public:
	FeaFullVector(int);
	~FeaFullVector();
	void operator=(const FeaFullVector&);
	FeaFullVector(const FeaFullVector&);

	double& operator()(int i){return data[i];};
	FeaFullVector& operator+=(FeaFullVector&);
	FeaFullVector& operator-=(FeaFullVector&);
	FeaFullVector& operator*=(double);
	FeaFullVector& operator+(FeaFullVector&);
	FeaFullVector& operator-(FeaFullVector&);
	double operator*(FeaFullVector& v);		// the tanspose of this vector times the given v
	FeaFullVector& operator*(double);

	void assemble(FeaFullVector&, int*);
	FeaFullVector& getSubVector(int, int);
	void zero();

	double oneNorm();	// calculates the one norm of the vector
	double twoNorm();	// calculates the two norm of the vector

	double oneNorm(double limit);
	double twoNorm(double limit);


	double infinityNorm();	// calculates the infinity norm of the vector
	void twoNormalize();	// divides the vector by the two norm
	void orthogonalize(FeaFullMatrix& orthoColVectors, int startCol, int endCol, FeaMatrix& wrt);	// orthogonalize the vector with the given columns wrt the given matrix
	void normalizeWrt(FeaMatrix& m);

	int getSize();
	
	void print();
	void printTerse();
   
	friend class FeaFullMatrix;
	friend class FeaColCompMatrix;
	friend class FeaBlockMatrix;
	friend class FeaSymmetricMatrix;
};


/****************************/
/*        FeaMatrix         */
/****************************/

class FeaMatrix {
	matrixType type;
public:
	FeaMatrix(matrixType t) { type = t; };
	virtual ~FeaMatrix(){};
	void operator=(const FeaMatrix& other) { type = other.type; };
	FeaMatrix(const FeaMatrix& other){ type = other.type; };
	int operator==(matrixType t){return(type == t);};

	virtual double& operator()(int, int)=0;

	virtual FeaMatrix& operator*(FeaMatrix&)=0;
	virtual FeaFullVector& operator*(FeaFullVector&)=0;
	virtual FeaMatrix& operator*(double)=0;
   
	virtual FeaMatrix& operator+=(FeaMatrix&);

	virtual FeaFullMatrix& getSubMatrix(int, int, int, int)=0;

   virtual FeaMatrix& tripleProduct(FeaFullMatrix&) = 0;
   virtual double tripleProduct(FeaFullVector&) = 0;

	virtual void assemble(FeaMatrix&, int*)=0;

	virtual void luDecomp()=0;
	virtual void luForBack(FeaFullVector&)=0;
	virtual void zero()=0;
	virtual double maxVal()=0;	// return the largest (absolute value) value in the matrix

    // returns the eigenvectors, and places the values in eigenValues,  number needed given as length of vector, does not destroy the matrix
    FeaFullMatrix& eigen(FeaFullVector& eigenValues, FeaMatrix& mass);

	virtual int nRow()=0;
	virtual int nCol()=0;
	virtual FeaMatrix& transpose()=0;
	virtual FeaMatrix& copy()=0;
	virtual void print()=0;
};

/****************************/
/*   FeaFullMatrix class    */
/****************************/
class FeaBlockMatrix;

class FeaFullMatrix : public FeaMatrix {
	int row;
	int col;
	double** data;
public:
	FeaFullMatrix(int,int);
	~FeaFullMatrix();
	void operator=(const FeaFullMatrix&);
	FeaFullMatrix(const FeaFullMatrix&);

	void print();
	double& operator()(int i, int j) {return data[i][j];};

	FeaMatrix& operator*(FeaMatrix&);
	FeaFullMatrix& operator*(FeaFullMatrix&);
	FeaBlockMatrix& operator*(FeaBlockMatrix&);
	FeaFullVector& operator*(FeaFullVector&);
	FeaMatrix& operator*(double);
	FeaFullMatrix& operator*=(double);
	FeaFullVector& transposeMultiply(FeaFullVector&);

	FeaFullMatrix& operator+(FeaFullMatrix&);
	FeaMatrix& operator+=(FeaMatrix&);
	FeaFullMatrix& operator+=(FeaFullMatrix&);

	FeaMatrix& tripleProduct(FeaFullMatrix&);
	double tripleProduct(FeaFullVector&);

	FeaMatrix& transpose();
	FeaFullMatrix& getSubMatrix(int, int, int, int);

	void assemble(FeaMatrix&, int*);
	void assemble(FeaFullMatrix&, int*);
    void assemble(FeaSymmetricMatrix&, int*);

	void luDecomp();
	void luForBack(FeaFullVector&);
	void MGSOrthoWrt(FeaFullMatrix& m);

	void zero();
	double maxVal();

	int nRow();
	int nCol();
	FeaMatrix& copy();
	friend class FeaFullVector;
	friend class FeaColCompMatrix;
	friend class FeaBlockMatrix;
	friend class FeaSymmetricMatrix;
   	friend class FeaSuperElementMatrix;

};

/****************************/
/*     FeaBlockMatrix       */
/****************************/

class FeaBlockMatrix : public FeaMatrix {
	int row;
	int col;
	int nRowBlocks;
	int nColBlocks;
	FeaFullMatrix*** data;
	int* rowEnd;
	int* colEnd;
public:
	FeaBlockMatrix(int,int); // initialize with number of row and column blocks
	~FeaBlockMatrix();
	void operator=(const FeaBlockMatrix&);
	FeaBlockMatrix(const FeaBlockMatrix&);
	void print();
	double& operator()(int, int);
	void addBlock(FeaFullMatrix&, int, int); // add a block with top left hand absolute corner


	//add by Hung-Ming
	void reSize(int,int);

	FeaMatrix& operator*(FeaMatrix&);
	FeaBlockMatrix& operator*(FeaBlockMatrix&);
	FeaBlockMatrix& operator*(FeaFullMatrix&);
	FeaFullVector& operator*(FeaFullVector&);
	FeaMatrix& operator*(double);

	FeaMatrix& operator+=(FeaMatrix&);
	FeaMatrix& transpose();
	FeaFullMatrix& getSubMatrix(int, int, int, int); //top left absolute postion and row and column length
	FeaFullMatrix& getBlock00();  // warning, creates an alias!!!!!
	FeaFullMatrix& getBlock(int, int);  // warning, creates an alias!!!!!

	void assemble(FeaMatrix&, int*);

    FeaMatrix& tripleProduct(FeaFullMatrix&);
	double tripleProduct(FeaFullVector&);

	void luDecomp();
	void luForBack(FeaFullVector&);
	void zero();
	double maxVal();

	int nRow();
	int nCol();
	int nRowBlock();
	int nColBlock();
	FeaMatrix& copy();

	friend class FeaFullVector;
	friend class FeaFullMatrix;
	friend class FeaColCompMatrix;
    friend class FeaSymmetricMatrix;
};

/****************************/
/*     FeaColCompMatrix     */
/****************************/

class FeaColCompMatrix : public FeaMatrix {
	int size;
	double* data;
	int* diag;
public:
	FeaColCompMatrix(int,int*);
	~FeaColCompMatrix();
	void operator=(const FeaColCompMatrix&);
	FeaColCompMatrix(const FeaColCompMatrix&);
	void print();

	double& operator()(int row, int col){return data[((col<row) ? diag[row] : diag[col]) - (abs(col-row))];};

	FeaMatrix& operator*(FeaMatrix&);
	FeaFullMatrix& operator*(FeaFullMatrix&);
	FeaFullVector& operator*(FeaFullVector&);
	FeaMatrix& operator*(double);

	FeaMatrix& operator+=(FeaMatrix&);
	FeaColCompMatrix& operator+=(FeaColCompMatrix&);

	FeaMatrix& transpose();

	FeaFullMatrix& getSubMatrix(int, int, int, int); //top left absolute postion and row and column length

	void assemble(FeaMatrix&, int*);
	void assemble(FeaFullMatrix&, int*);
    void assemble(FeaSymmetricMatrix&, int*);

    FeaMatrix& tripleProduct(FeaFullMatrix&);
	double tripleProduct(FeaFullVector&);

	void luDecomp();
	void luForBack(FeaFullVector&);

	void zero();
	int nRow();
	int nCol();

	FeaMatrix& copy();
	double maxVal();

	friend class FeaFullVector;
	friend class FeaFullMatrix;
	friend class FeaBlockMatrix;
    friend class FeaSymmetricMatrix;
};

/****************************/
/*   FeaSymmetricMatrix     */
/****************************/

class FeaSymmetricMatrix : public FeaMatrix {
	int size;
	double* data;
	int* diag;
public:
	FeaSymmetricMatrix(int);
	~FeaSymmetricMatrix();
	void operator=(const FeaSymmetricMatrix&);
	FeaSymmetricMatrix(const FeaSymmetricMatrix&);
	void print();

	double& operator()(int row, int col){return data[((col<row) ? diag[row] : diag[col]) - (abs(col-row))];};

	FeaMatrix& operator*(FeaMatrix&);
	FeaFullMatrix& operator*(FeaFullMatrix&);
	FeaFullVector& operator*(FeaFullVector&);
	FeaMatrix& operator*(double);

	FeaMatrix& operator+=(FeaMatrix&);
	FeaSymmetricMatrix& operator+=(FeaSymmetricMatrix&);

	FeaMatrix& transpose();

	FeaFullMatrix& getSubMatrix(int, int, int, int); //top left absolute postion and row and column length

	void assemble(FeaMatrix&, int*);
	void assemble(FeaFullMatrix&, int*);
    void assemble(FeaSymmetricMatrix&, int*);

    FeaMatrix& tripleProduct(FeaFullMatrix&);
	double tripleProduct(FeaFullVector&);

	void luDecomp();
	void luForBack(FeaFullVector&);

	void zero();
	int nRow();
	int nCol();

	FeaMatrix& copy();
	double maxVal();

	friend class FeaFullVector;
	friend class FeaFullMatrix;
	friend class FeaBlockMatrix;
	friend class FeaColCompMatrix;
	friend class FeaSuperElementMatrix;
};


